import NomPa.Derived
import NomPa.Derived.NaPa
import NomPa.Encodings.NominalTypes
open import NomPa.Record
open import Function.NP
open import Data.Product.Extras using (_×_;∃;_,_;proj₁;proj₂)
open import Data.Sum
open import Data.Nat
open import Data.Empty
open import Data.Maybe.NP
open import Data.Unit using (⊤)
open import Relation.Binary.PropositionalEquality

module NomPa.Encodings.AlphaCaml (nomPa : NomPa) where

open NomPa nomPa
open NomPa.Derived nomPa
open NomPa.Derived.NaPa nomPa
private
  module NomTypes = NomPa.Encodings.NominalTypes nomPa
open NomTypes public
  using (_×ᵉ_; _⊎ᵉ_; _→ᵉ_; _↦ᵉ_; ∀ᵉ; Kᵉ; NomSet; 𝔼; Nameᵉ; Listᵉ; Maybeᵉ)

ℙ : Set₁
ℙ = (α β : World) (Op : World → World) → Set

Binderᵖ : ℙ
Binderᵖ _ _ Op = ∃[ b ](Op ≡ _◅_ b)

-- de Bruijn version!
Binderᴰᵖ : ℙ
Binderᴰᵖ _ _ Op = Op ≡ _↑1

Outer : NomSet → ℙ
Outer E α _ Op = E α × Op ≡ id

Inner : NomSet → ℙ
Inner E _ β Op = E β × Op ≡ id

-- Kᵖ is called neutral in Cαml
Kᵖ : Set → ℙ
Kᵖ A _ _ Op = A × Op ≡ id
-- Kᵖ A ≡ Inner (Kᵉ A) ≡ Outer (Kᵉ A)

<_> : ℙ → NomSet
< P > α = ∃[ Op ](P α (Op α) Op)

∀ᵖ : ℙ → Set
∀ᵖ P = ∀ {α β Op} → P α β Op

infixr 0 _→ᵖ_ _↦ᵖ_
_→ᵖ_ : ℙ → ℙ → ℙ
_→ᵖ_ P₁ P₂ α β Op = P₁ α β Op → P₂ α β Op

_↦ᵖ_ : ℙ → ℙ → Set
_↦ᵖ_ P₁ P₂ = ∀ᵖ (P₁ →ᵖ P₂)

infixr 4 _,_
infixr 2 _×ᵖ_
data _×ᵖ_ (P₁ P₂ : ℙ) : ℙ where
  _,_ : ∀ {Op₁ Op₂ α β} → P₁ α β Op₁ → P₂ α β Op₂ → (P₁ ×ᵖ P₂) α β (Op₂ ∘ Op₁)

infixr 1 _⊎ᵖ_
_⊎ᵖ_ : ℙ → ℙ → ℙ
(P₁ ⊎ᵖ P₂) α β Op = P₁ α β Op ⊎ P₂ α β Op

data Listᵖ (P : ℙ) : ℙ where
  -- Kᵖ ⊤ provides the Op ≡ id
  [] : Kᵖ ⊤ ↦ᵖ Listᵖ P
  -- ×ᵖ instead of →ᵖ provides the Op ≡ Op₁ ∘ Op₂
  ∷  : (P ×ᵖ Listᵖ P) ↦ᵖ Listᵖ P

open import Data.List
open import Data.Maybe

module FreeVars where
  open NomTypes.FreeVars public

  record Fv′ᵖ α β (Op : World → World) : Set where
    constructor mk
    field
      fvO : List (Name α)
      fvI : List (Name β)
      rmP : ∀ {γ} → List (Name (Op γ)) → List (Name γ)

  Fvᵖ : ℙ → Set
  Fvᵖ P = P ↦ᵖ Fv′ᵖ

  fv<> : ∀ {P} → Fvᵖ P → Fv < P >
  fv<> fvᵖ (_ , p) with fvᵖ p
  ... | mk fvO fvI rmP = fvO ++ rmP fvI

  abstract -- only here for debugging purposes
    fvᵖdummy : ∀ {A : Set} {α β Op} → A → Fv′ᵖ α β Op
    fvᵖdummy _ = mk [] [] (const [])

  fvᵖMap : ∀ {P Q} → (Q ↦ᵖ P) → Fvᵖ P → Fvᵖ Q
  fvᵖMap f fvP = fvP ∘ f

  fvBinderᵖ : Fvᵖ Binderᵖ
  fvBinderᵖ (b , refl) = mk [] [] (rm b)

  fvBinderᴰᵖ : Fvᵖ Binderᴰᵖ
  fvBinderᴰᵖ refl = mk [] [] rm₀

  _++ᵖ_ : ∀ {α β Op₁ Op₂} → Fv′ᵖ α β Op₁ → Fv′ᵖ α β Op₂ → Fv′ᵖ α β (Op₂ ∘ Op₁)
  mk fvO₁ fvI₁ rmP₁ ++ᵖ mk fvO₂ fvI₂ rmP₂ = mk (fvO₁ ++ fvO₂) (fvI₁ ++ fvI₂) (rmP₁ ∘ rmP₂)

  fv×ᵖ : ∀ {P₁ P₂} → Fvᵖ P₁ → Fvᵖ P₂ → Fvᵖ (P₁ ×ᵖ P₂)
  fv×ᵖ fv₁ fv₂ (p₁ , p₂) = fv₁ p₁ ++ᵖ fv₂ p₂

  fv⊎ᵖ : ∀ {P₁ P₂} → Fvᵖ P₁ → Fvᵖ P₂ → Fvᵖ (P₁ ⊎ᵖ P₂)
  fv⊎ᵖ fv₁ fv₂ = [ fv₁ , fv₂ ]′

  fvInner : ∀ {E} → Fv E → Fvᵖ (Inner E)
  fvInner fvE (x , refl) = mk [] (fvE x) id

  fvOuter : ∀ {E} → Fv E → Fvᵖ (Outer E)
  fvOuter fvE (x , refl) = mk (fvE x) [] id

  fvKᵖ : ∀ {A} → Fvᵖ (Kᵖ A)
  fvKᵖ (_ , refl) = mk [] [] id

  fvListᵖ : ∀ {P : ℙ} → Fvᵖ P → Fvᵖ (Listᵖ P)
  fvListᵖ fvP ([] k) = fvKᵖ k
  fvListᵖ fvP (∷ (x , xs)) = fvP x ++ᵖ fvListᵖ fvP xs

module LC-Example where
  data Tm : NomSet where
    var  : Nameᵉ ↦ᵉ Tm
    app  : (Tm ×ᵉ Tm) ↦ᵉ Tm
    lam  : < Binderᵖ ×ᵖ Inner Tm > ↦ᵉ Tm
    let′ : < Binderᵖ ×ᵖ Outer Tm ×ᵖ Inner Tm > ↦ᵉ Tm

  module Ctors where
    _·_ : Tm ↦ᵉ Tm →ᵉ Tm
    _·_ t u = app (t , u)

    ƛ : ∀ {α} b → Tm (b ◅ α) → Tm α
    ƛ b t = lam (_ , (b , refl) , (t , refl))

    Let : ∀ {α} b → Tm α → Tm (b ◅ α) → Tm α
    Let b t u = let′ (_ , (b , refl) , ((t , refl) , (u , refl)))

  module Terms where
    open Ctors
    idTm : Tm ø
    idTm = ƛ (0 ᴮ) (var (0 ᴺ))
    apTm : Tm ø
    apTm = ƛ (0 ᴮ) (ƛ (1 ᴮ) (var (0 ᴺ¹) · var (1 ᴺ)))

module LC-Example2 where
  TmF : NomSet → NomSet
  TmF Tm =  {-var-} Nameᵉ
         ⊎ᵉ {-app-} (Tm ×ᵉ Tm)
         ⊎ᵉ {-lam-} < Binderᵖ ×ᵖ Inner Tm >
         ⊎ᵉ {-let-} < Binderᵖ ×ᵖ Outer Tm ×ᵖ Inner Tm >

{-
  postulate
    μᵉ : (NomSet → NomSet) → NomSet
    roll   : ∀ {F} → F (μᵉ F) ↦ᵉ μᵉ F
    unroll : ∀ {F} → μᵉ F ↦ᵉ F (μᵉ F)
  Tm = μᵉ TmF
-}

  data Tm : NomSet where
    roll : TmF Tm ↦ᵉ Tm
  unrollTm : Tm ↦ᵉ TmF Tm
  unrollTm (roll x) = x

  module Ctors where
    var : Nameᵉ ↦ᵉ Tm
    var = roll ∘′ inj₁

    app : Tm ↦ᵉ Tm →ᵉ Tm
    app t u = roll (inj₂ (inj₁ (t , u)))

    lam : ∀ {α} b → Tm (b ◅ α) → Tm α
    lam b t = roll (inj₂ (inj₂ (inj₁ (_ , (b , refl) , (t , refl)))))

  module Terms where
    open Ctors
    idTm : Tm ø
    idTm = lam (0 ᴮ) (var (0 ᴺ))
    apTm : Tm ø
    apTm = lam (0 ᴮ) (lam (1 ᴮ) (app (var (0 ᴺ¹)) (var (1 ᴺ))))

module ML-Example where
  Tag = ℕ

  mutual
    data Tm : NomSet where
      var   : Nameᵉ ↦ᵉ Tm
      app   : Tm ×ᵉ Tm ↦ᵉ Tm
      lam   : < Binderᵖ ×ᵖ Inner Tm > ↦ᵉ Tm
      ctor  : Kᵉ Tag ×ᵉ Listᵉ Tm ↦ᵉ Tm
      let′  : < Binderᵖ ×ᵖ Outer Tm ×ᵖ Inner Tm > ↦ᵉ Tm
      rec   : < Listᵖ (Binderᵖ ×ᵖ Inner Tm) ×ᵖ Inner Tm > ↦ᵉ Tm
      match : Tm ×ᵉ Listᵉ Branch ↦ᵉ Tm

    Branch : NomSet
    Branch = < Pa ×ᵖ Inner (Maybeᵉ Tm) ×ᵖ Inner Tm >

    data Pa : ℙ where
      wildcard : Kᵖ ⊤ ↦ᵖ Pa
      binder   : Binderᵖ ↦ᵖ Pa
      pair     : Pa ×ᵖ Pa ↦ᵖ Pa
      ctor     : Kᵖ Tag ×ᵖ Listᵖ Pa ↦ᵖ Pa
      as       : Pa ×ᵖ Binderᵖ ↦ᵖ Pa

module ML-Example2 where
  Tag = ℕ

  module Types (Tm : NomSet) (Pa : ℙ) where
    mutual
      TmF : NomSet
      TmF = {-var-} Nameᵉ
          ⊎ᵉ {-app-}   Tm ×ᵉ Tm
          ⊎ᵉ {-lam-}   < Binderᵖ ×ᵖ Inner Tm >
          ⊎ᵉ {-ctor-}  Kᵉ Tag ×ᵉ Listᵉ Tm
          ⊎ᵉ {-let-}   < Binderᵖ ×ᵖ Outer Tm ×ᵖ Inner Tm >
          ⊎ᵉ {-rec-}   < Listᵖ (Binderᵖ ×ᵖ Inner Tm) ×ᵖ Inner Tm >
          ⊎ᵉ {-match-} Tm ×ᵉ Listᵉ Branch

      Branch : NomSet
      Branch = < Pa ×ᵖ Inner (Maybeᵉ Tm) ×ᵖ Inner Tm >

      PaF : ℙ
      PaF =  {-wildcard-} Kᵖ ⊤
          ⊎ᵖ {-binder-}   Binderᵖ
          ⊎ᵖ {-pair-}     Pa ×ᵖ Pa
          ⊎ᵖ {-ctor-}     Kᵖ Tag ×ᵖ Listᵖ Pa
          ⊎ᵖ {-as-}       Pa ×ᵖ Binderᵖ

  mutual
    data Tm : NomSet where
      roll : Types.TmF Tm Pa ↦ᵉ Tm
    data Pa : ℙ where
      roll : Types.PaF Tm Pa ↦ᵖ Pa

  open Types Tm Pa

  unrollTm : Tm ↦ᵉ TmF
  unrollTm (roll x) = x
  unrollPa : Pa ↦ᵖ PaF
  unrollPa (roll x) = x

  open FreeVars
  mutual
    fvTmF : ℕ → Fv TmF
    fvTmF n = fv⊎ᵉ fvNameᵉ
             (fv⊎ᵉ (fv×ᵉ (fvTm n) (fvTm n))
             (fv⊎ᵉ (fv<> (fv×ᵖ fvBinderᵖ (fvInner (fvTm n))))
             (fv⊎ᵉ (fv×ᵉ fvKᵉ (fvListᵉ (fvTm n)))
             (fv⊎ᵉ (fv<> (fv×ᵖ fvBinderᵖ (fv×ᵖ (fvOuter (fvTm n)) (fvInner (fvTm n)))))
             (fv⊎ᵉ (fv<> (fv×ᵖ (fvListᵖ (fv×ᵖ fvBinderᵖ (fvInner (fvTm n)))) (fvInner (fvTm n))))
             (fv×ᵉ (fvTm n) (fvListᵉ (fvBranch n))))))))

    fvTm : ℕ → Fv Tm
    fvTm zero    = fvDummy
    fvTm (suc n) = fvTmF n ∘ unrollTm

    fvBranch : ℕ → Fv Branch
    fvBranch n = fv<> (fv×ᵖ (fvPa n) (fv×ᵖ (fvInner (fvMaybeᵉ (fvTm n))) (fvInner (fvTm n))))

    fvPa : ℕ → Fvᵖ Pa
    fvPa zero    = fvᵖdummy
    fvPa (suc n) = fvᵖMap unrollPa (fvPaF n)

    fvPaF : ℕ → Fvᵖ PaF
    fvPaF n = fv⊎ᵖ fvKᵖ
             (fv⊎ᵖ fvBinderᵖ
             (fv⊎ᵖ (fv×ᵖ (fvPa n) (fvPa n))
             (fv⊎ᵖ (fv×ᵖ fvKᵖ (fvListᵖ (fvPa n)))
                    (fv×ᵖ (fvPa n) fvBinderᵖ))))

  module Ctors where
    var : Nameᵉ ↦ᵉ Tm
    var = roll ∘′ inj₁

    app : Tm ↦ᵉ Tm →ᵉ Tm
    app t u = roll (inj₂ (inj₁ (t , u)))

    lam : ∀ {α} b → Tm (b ◅ α) → Tm α
    lam b t = roll (inj₂ (inj₂ (inj₁ (_ , (b , refl) , t , refl))))

    unLam : ∀ {α} → Tm α →? (∃[ b ](Tm (b ◅ α)))
    unLam (roll (inj₂ (inj₂ (inj₁ (._ , (b , refl) , t , refl))))) = just (b , t)
    unLam _ = nothing

  module Terms where
    open Ctors

    idTm : Tm ø
    idTm = lam (0 ᴮ) (var (0 ᴺ))

    apTm : Tm ø
    apTm = lam (0 ᴮ) (lam (1 ᴮ) (app (var (0 ᴺ¹)) (var (1 ᴺ))))
