open import NomPa.Record
import NomPa.Encodings.AlphaCaml
import NomPa.Derived
open import Data.List
open import Function.NP
open import Data.Product.Extras using (_×_;∃;_,_)
open import Relation.Binary.PropositionalEquality
import Data.Indexed

module NomPa.Encodings.BindersUnbound (nomPa : NomPa) where

open NomPa nomPa
open NomPa.Derived nomPa

module Cαml = NomPa.Encodings.AlphaCaml nomPa

open Cαml using (Inner; Outer; <_>)

-- All these are re-exported
open Cαml public using (NomSet; Nameᵉ; _×ᵉ_; _⊎ᵉ_; Kᵉ; _→ᵉ_; _↦ᵉ_; ∀ᵉ;
                        ℙ; Binderᵖ; _×ᵖ_; _⊎ᵖ_; Kᵖ; _→ᵖ_; _↦ᵖ_; ∀ᵖ;
                        Listᵉ; Maybeᵉ; Listᵖ)

Bind : ℙ → NomSet → NomSet
Bind P E α = ∃[ Op ](P α (Op α) Op × E (Op α))
        -- ≡ ∃[ Op ](∃[ Op₁ ](∃[ Op₂ ](Op ≡ Op₂ ∘ Op₁ × P α (Op α) Op₁ × E (Op α) × Op₂ ≡ id)))
        -- ≡ ∃[ Op ](∃[ Op₁ ](∃[ Op₂ ](Op ≡ Op₂ ∘ Op₁ × P α (Op α) Op₁ × Inner E α (Op α) Op₂)))
        -- ≅ ∃[ Op ]((P ×ᵖ Inner E) α (Op α) Op)
        -- ≡ < P ×ᵖ Inner E > α

Embed : NomSet → ℙ
Embed = Outer

{- from «Binders Unbound»
Rebind P₁ P₂ acts like the pattern type
(P₁ , P₂), except that P₁ also scopes over P₂, so the binders in P₁
may be referred to by terms embedded within P₂. (The fact that P₁
scopes over P₂ in this way has no effect on the pattern portion of
P₂.) -}
infixr 4 _,_
data Rebind (P₁ P₂ : ℙ) : ℙ where
  _,_ : ∀ {Op₁ Op₂ α β} → P₁ α β Op₁ → P₂ (Op₁ α) β Op₂ → Rebind P₁ P₂ α β (Op₂ ∘ Op₁)

{- from «Binders Unbound»
In Rec P, names in the pattern P scope recursively
over any terms embedded in P itself. However, Rec P itself is
also a pattern, so names in P also scope externally over any term
that binds Rec P . Intuitively, Rec is just a “recursive” version of
Rebind.
-}
Rec : ℙ → ℙ
Rec P α β Op = P (Op α) β Op

module FreeVars where
  -- hum there seems to be an Agda bug behind this...
  -- If I import Fv from here, the Fv is still parameterized by NomPa
  open Cαml.FreeVars public using ({-Fv;-} fv×ᵉ; fv⊎ᵉ; fvKᵉ; fvNameᵉ; mk;
                                   Fv′ᵖ; {-Fvᵖ;-} _++ᵖ_; fv×ᵖ; fv⊎ᵖ; fvBinderᵖ; fvInner; fvOuter; fvKᵖ)

  Fv : NomSet → Set
  Fv E = E ↦ᵉ Listᵉ Nameᵉ

  Fvᵖ : ℙ → Set
  Fvᵖ P = P ↦ᵖ Fv′ᵖ

  fvBind : ∀ {P} {E : NomSet} → Fvᵖ P → Fv E → Fv (Bind P E)
  fvBind fvP fvE (_ , p , e) with fvP p
  ... | mk fvO fvI rmP = fvO ++ rmP (fvI ++ fvE e)

  fvEmbed : ∀ {E} → Fv E → Fvᵖ (Embed E)
  fvEmbed = Cαml.FreeVars.fvOuter

  fvRec : ∀ {P} → Fvᵖ P → Fvᵖ (Rec P)
  fvRec fvP p with fvP p
  ... | mk fvO fvI rmP = mk (rmP fvO) fvI rmP

  fvRebind : ∀ {P₁ P₂} → Fvᵖ P₁ → Fvᵖ P₂ → Fvᵖ (Rebind P₁ P₂)
  fvRebind fvP₁ fvP₂ (p₁ , p₂) with fvP₁ p₁ | fvP₂ p₂
  ... | mk fvO₁ fvI₁ rmP₁ | mk fvO₂ fvI₂ rmP₂ = mk (fvO₁ ++ rmP₁ fvO₂) (fvI₁ ++ fvI₂) (rmP₁ ∘ rmP₂)
