open import NomPa.Record
import NomPa.Derived
import NomPa.Derived.NaPa
import NomPa.Traverse
import Data.Indexed
open import Function.NP
open import Data.Product using (_×_;∃;_,_)
open import Data.Sum using (_⊎_; inj₁; inj₂)
open import Data.List
open import Data.Maybe
open import Data.Sum
open import Category.Applicative renaming (module RawApplicative to Applicative; RawApplicative to Applicative)

module NomPa.Encodings.NominalTypes (nomPa : NomPa) where

open NomPa nomPa
open NomPa.Derived nomPa
open NomPa.Derived.NaPa nomPa using (SynAbsᴰ)
open NomPa.Traverse nomPa
open Data.Indexed {_} {World} public
  renaming (_|×|_ to _×ᵉ_; _|⊎|_ to _⊎ᵉ_; _|→|_ to _→ᵉ_; _|↦|_ to _↦ᵉ_;
            |K| to Kᵉ; |∀| to ∀ᵉ)

NomSet : Set₁
NomSet = |Set| _

𝔼 = NomSet

Nameᵉ : NomSet
Nameᵉ = Name

Absᵉ : NomSet → NomSet
Absᵉ = SynAbsᴺ

Absᴰᵉ : NomSet → NomSet
Absᴰᵉ = SynAbsᴰ

Empty : NomSet → Set
Empty E = E ø

Listᵉ : NomSet → NomSet
Listᵉ = _∘_ List

Maybeᵉ : NomSet → NomSet
Maybeᵉ = _∘_ Maybe

module FreeVars where
  Fv : NomSet → Set
  Fv E = E ↦ᵉ Listᵉ Nameᵉ

  fv×ᵉ : ∀ {E₁ E₂} → Fv E₁ → Fv E₂ → Fv (E₁ ×ᵉ E₂)
  fv×ᵉ fv₁ fv₂ (x , y) = fv₁ x ++ fv₂ y

  fv⊎ᵉ : ∀ {E₁ E₂} → Fv E₁ → Fv E₂ → Fv (E₁ ⊎ᵉ E₂)
  fv⊎ᵉ fv₁ fv₂ = [ fv₁ , fv₂ ]′

  fvKᵉ : ∀ {A} → Fv (Kᵉ A)
  fvKᵉ _ = []

  fvNameᵉ : Fv Nameᵉ
  fvNameᵉ = [_]

  fvListᵉ : ∀ {E} → Fv E → Fv (Listᵉ E)
  fvListᵉ _   []       = []
  fvListᵉ fvE (x ∷ xs) = fvE x ++ fvListᵉ fvE xs

  fvMaybeᵉ : ∀ {E} → Fv E → Fv (Maybeᵉ E)
  fvMaybeᵉ fvE = maybe fvE []

  abstract -- only here for debugging purposes
    fvDummy : ∀ {A B : Set} → A → List B
    fvDummy = const []

  fvMap : ∀ {E₁ E₂} → (E₂ ↦ᵉ E₁) → Fv E₁ → Fv E₂
  fvMap f fvE₁ = fvE₁ ∘ f
