open import NomPa.Record
open import Function.NP
open import Data.List
open import Data.Sum
open import Data.Product
open import Data.Maybe.NP as Maybe
open import Data.Bool
open import Data.Nat
import Data.Indexed
open import Data.String as String hiding (_++_)
open import Category.Applicative
            renaming (RawApplicative to Applicative; module RawApplicative to Applicative)
open import Relation.Binary.PropositionalEquality as ≡
import NomPa.Derived
import NomPa.Derived.NaPa
import NomPa.Traverse
open import NomPa.Examples.Raw
open import NomPa.Examples.Raw.Parser
import NomPa.Examples.LC

module NomPa.Examples.LC.DbLevels (nomPa : NomPa)
                                  (Cst : Set)
                                  (_==ᶜ_ : Cmp Cst)
                                  (ℕ⇒Cst : ℕ → Cst)
                                  (Cst⇒ℕ : Cst → ℕ) where

open NomPa nomPa
open NomPa.Derived nomPa
open NomPa.Derived.NaPa nomPa hiding (module Env)
open NomPa.Traverse nomPa
open NomPa.Examples.LC nomPa using (V; ƛ; _·_; Let; `_) renaming (Tm to Tmᴺ; showTmø to showTmᴺ)
open Data.Indexed {_} {World}

data Tmᴸ b α : Set where
  V   : (x : Name α) → Tmᴸ b α
  ƛ   : (t : Tmᴸ (sucᴮ b) (b ◅ α)) → Tmᴸ b α
  _·_ : (t u : Tmᴸ b α) → Tmᴸ b α
  Let : (t : Tmᴸ b α) (u : Tmᴸ (sucᴮ b) (b ◅ α)) → Tmᴸ b α
  `_  : (κ : Cst) → Tmᴸ b α

fv : ∀ {b α} → Tmᴸ b α → List (Name α)
fv (V x)        = [ x ]
fv (fct · arg)  = fv fct ++ fv arg
fv (ƛ t)        = rm _ (fv t)
fv (Let t u)    = fv t ++ rm _ (fv u)
fv (` _)        = []

module CoerceTmByHand where
  coerceTmᴸ : ∀ {α β b} → (α ⊆ β) → Tmᴸ b α → Tmᴸ b β
  coerceTmᴸ α⊆β (V x) = V (x ⟨-because α⊆β -⟩)
  coerceTmᴸ α⊆β (t · u) = coerceTmᴸ α⊆β t · coerceTmᴸ α⊆β u
  coerceTmᴸ α⊆β (ƛ t) = ƛ (coerceTmᴸ (⊆-◅ _ α⊆β) t)
  coerceTmᴸ α⊆β (Let t u) = Let (coerceTmᴸ α⊆β t) (coerceTmᴸ (⊆-◅ _ α⊆β) u)
  coerceTmᴸ _    (` κ) = ` κ

-- generalized further by AddSubtractTmByHand
module AddTmByHand k where
  Env : EnvTypeᴸ
  Env b₁ b₂ α β = (α +ᵂ k) ⊆ β × (b₁ +ᴮ k) ≡ b₂

  extEnv : ExtEnvᴸ Env
  extEnv (pf , ≡.refl) = ⊆-trans (⊆-dist-+-◅ k) (⊆-◅ _ pf) , +ᴮ-dist1 k

  trName : TrNameᴸ Env (const Name)
  trName (pf , _) = coerceᴺ pf ∘ addᴺ k

  trTmᴸ : ∀ {α β b₁ b₂} → Env b₁ b₂ α β → Tmᴸ b₁ α → Tmᴸ b₂ β
  trTmᴸ Δ (V x)     = V (trName Δ x)
  trTmᴸ Δ (t · u)   = trTmᴸ Δ t · trTmᴸ Δ u
  trTmᴸ Δ (ƛ t)     = ƛ (trTmᴸ (extEnv Δ) t)
  trTmᴸ Δ (Let t u) = Let (trTmᴸ Δ t) (trTmᴸ (extEnv Δ) u)
  trTmᴸ Δ (` κ)     = ` κ

  addTmᴸ : ∀ {α b} → Tmᴸ b α → Tmᴸ (b +ᴮ k) (α +ᵂ k)
  addTmᴸ = trTmᴸ (⊆-refl , ≡.refl)

module TraverseTmᴸ {Envᴸ} (kitᴸ : TrKitᴸ Envᴸ Tmᴸ) where
  open TrKitᴸ kitᴸ

  trTmᴸ : ∀ {α β b₁ b₂} → Envᴸ b₁ b₂ α β → Tmᴸ b₁ α → Tmᴸ b₂ β
  trTmᴸ Δ (V x)     = trNameᴸ Δ x
  trTmᴸ Δ (t · u)   = trTmᴸ Δ t · trTmᴸ Δ u
  trTmᴸ Δ (ƛ t)     = ƛ (trTmᴸ (extEnvᴸ Δ) t)
  trTmᴸ Δ (Let t u) = Let (trTmᴸ Δ t) (trTmᴸ (extEnvᴸ Δ) u)
  trTmᴸ _ (` κ)     = ` κ

module TraverseNameTmᴸ {Envᴸ} (kitᴸ : TrKitᴸ Envᴸ (const Name)) where
  open TrKitᴸ kitᴸ

  trTmᴸ : ∀ {α β b₁ b₂} → Envᴸ b₁ b₂ α β → Tmᴸ b₁ α → Tmᴸ b₂ β
  trTmᴸ Δ (V x)     = V (trNameᴸ Δ x)
  trTmᴸ Δ (t · u)   = trTmᴸ Δ t · trTmᴸ Δ u
  trTmᴸ Δ (ƛ t)     = ƛ (trTmᴸ (extEnvᴸ Δ) t)
  trTmᴸ Δ (Let t u) = Let (trTmᴸ Δ t) (trTmᴸ (extEnvᴸ Δ) u)
  trTmᴸ _ (` κ)     = ` κ

module TraverseANameTmᴸ {E} (E-app : Applicative E)
                        {Envᴸ} (kitᴸ : TrKitᴸ Envᴸ (const (E ∘ Name))) where
  open Applicative E-app
  open TrKitᴸ kitᴸ

  trTmᴸ : ∀ {α β b₁ b₂} → Envᴸ b₁ b₂ α β → Tmᴸ b₁ α → E (Tmᴸ b₂ β)
  trTmᴸ Δ (V x)     = pure V ⊛ trNameᴸ Δ x
  trTmᴸ Δ (t · u)   = pure _·_ ⊛ trTmᴸ Δ t ⊛ trTmᴸ Δ u
  trTmᴸ Δ (ƛ t)     = pure ƛ ⊛ trTmᴸ (extEnvᴸ Δ) t
  trTmᴸ Δ (Let t u) = pure Let ⊛ trTmᴸ Δ t ⊛ trTmᴸ (extEnvᴸ Δ) u
  trTmᴸ _ (` κ)     = pure $ ` κ

module TraverseATmᴸ {E} (E-app : Applicative E)
                    {Envᴸ} (kitᴸ : TrKitᴸ Envᴸ (λ b α → E (Tmᴸ b α))) where
  open Applicative E-app
  open TrKitᴸ kitᴸ

  trTmᴸ : ∀ {α β b₁ b₂} → Envᴸ b₁ b₂ α β → Tmᴸ b₁ α → E (Tmᴸ b₂ β)
  trTmᴸ Δ (V x)     = trNameᴸ Δ x
  trTmᴸ Δ (t · u)   = pure _·_ ⊛ trTmᴸ Δ t ⊛ trTmᴸ Δ u
  trTmᴸ Δ (ƛ t)     = pure ƛ ⊛ trTmᴸ (extEnvᴸ Δ) t
  trTmᴸ Δ (Let t u) = pure Let ⊛ trTmᴸ Δ t ⊛ trTmᴸ (extEnvᴸ Δ) u
  trTmᴸ _ (` κ)     = pure $ ` κ

addSubtractTmᴸ : ∀ k₁ k₂
                   {α β b₁ b₂} → (b₁ +ᴮ k₁) ≡ (b₂ +ᴮ k₂)
                               → (α +ᵂ k₁) ⊆ (β +ᵂ k₂)
                               → Tmᴸ b₁ α → Tmᴸ b₂ β
addSubtractTmᴸ k₁ k₂ eq pf
  = TraverseNameTmᴸ.trTmᴸ (AddSubtractᴸ.kitᴸ k₁ k₂) (eq , pf)

addTmᴸ : ∀ {α b} k → Tmᴸ b α → Tmᴸ (b +ᴮ k) (α +ᵂ k)
addTmᴸ k = addSubtractTmᴸ k 0 ≡.refl ⊆-refl

shiftTmᴸ′ : ∀ {b₁ b₂ α β} k → b₁ +ᴮ k ≡ b₂ → (α +ᵂ k) ⊆ β → Tmᴸ b₁ α → Tmᴸ b₂ β
shiftTmᴸ′ k pf₁ pf₂ = addSubtractTmᴸ k 0 pf₁ pf₂

shiftTmᴸ : ∀ {α β b} k → (α +ᵂ k) ⊆ β → Tmᴸ b α → Tmᴸ (b +ᴮ k) β
shiftTmᴸ k = shiftTmᴸ′ k ≡.refl

shiftøTmᴸ : ∀ {α} k → Tmᴸ (0 ᴮ) ø → Tmᴸ (k ᴮ) α
shiftøTmᴸ k = shiftTmᴸ′ k (0ᴮ-+ᴮ-id k) (⊆-ø+ k)

subtractTmᴸ : ∀ {α b} k → Tmᴸ (b +ᴮ k) (α +ᵂ k) → Tmᴸ b α
subtractTmᴸ k = addSubtractTmᴸ 0 k ≡.refl ⊆-refl

coerceTmᴸ : ∀ {b α β} → α ⊆ β → Tmᴸ b α → Tmᴸ b β
coerceTmᴸ = shiftTmᴸ 0

coerceøTmᴸ : ∀ {b α} → Tmᴸ b ø → Tmᴸ b α
coerceøTmᴸ = coerceTmᴸ ⊆-ø

rerootTmᴸ : ∀ {α} k₁ k₂ → Tmᴸ (k₁ ᴮ) ø → Tmᴸ (k₂ ᴮ) α
rerootTmᴸ k₁ k₂ = addSubtractTmᴸ k₂ k₁ (+ᴮ-comm k₁ k₂) (⊆-ø+ k₂)

renameTmᴸ : ∀ {α β b₁ b₂} → b₂ # β → (Name α → Name β) → Tmᴸ b₁ α → Tmᴸ b₂ β
renameTmᴸ b₂#β trName = TraverseNameTmᴸ.trTmᴸ renameKitᴸ (trName , b₂#β)

protectedAddTmᴸ : ∀ {α b₁ b₂} k → b₂ # (α +ᵂ k) → Tmᴸ b₁ α → Tmᴸ b₂ (α +ᵂ k)
protectedAddTmᴸ k pf = renameTmᴸ pf (addᴺ k)

-- This function behave like addTmᴸ but protectedAddTmᴸ track separatedely
-- the bound and the free variables. This gives more flexibilty since one
-- can choose to move only the bound or only the free variables. However
-- addTmᴸ is faster and does not require a b#α
-- protectedAddTmᴸ′ : ∀ {α b} k → b # α → Tmᴸ b α → Tmᴸ (b +ᴮ k) (α +ᵂ k)
-- protectedAddTmᴸ′ k pf = protectedAddTmᴸ k (#-+ k)

protectedSubtractTmᴸ : ∀ {α b₁ b₂} k → b₂ # α → Tmᴸ b₁ (α +ᵂ k) → Tmᴸ b₂ α
protectedSubtractTmᴸ k pf = renameTmᴸ pf (subtractᴺ k)

renameATmᴸ : ∀ {E} (E-app : Applicative E)
               {α β b₁ b₂} → b₂ # β → (Name α → E (Name β)) → Tmᴸ b₁ α → E (Tmᴸ b₂ β)
renameATmᴸ E-app b₂#β trNameᴸ =
  TraverseANameTmᴸ.trTmᴸ E-app (renameAKitᴸ E-app) (trNameᴸ , b₂#β)

renameTmᴸ? : ∀ {α β b₁ b₂} → b₂ # β → (Name α →? Name β) → Tmᴸ b₁ α →? (Tmᴸ b₂ β)
renameTmᴸ? = renameATmᴸ Maybe.applicative

closeTmᴸ? : ∀ {b α} → Tmᴸ b α →? Tmᴸ b ø
closeTmᴸ? = renameTmᴸ? (_ #ø) (const nothing)

⊆-dist-+-◅… : ∀ n → (n ◅…′ ø) +1 ⊆ n ᴮ ◅ n ◅…′ ø
⊆-dist-+-◅… zero = ⊆-ø+ 1
⊆-dist-+-◅… (suc n) = ⊆-trans (⊆-dist-+-◅ 1) (⊆-◅ _ (⊆-dist-+-◅… n))

shiftTmᴸ1′◅… : ∀ {α b} n → b ≡ n ᴮ → α ≡ n ◅…′ ø → Tmᴸ b α → Tmᴸ (sucᴮ b) (b ◅ α)
shiftTmᴸ1′◅… n ≡.refl ≡.refl = shiftTmᴸ′ 1 ≡.refl (⊆-dist-+-◅… n)

importTm : ∀ {b α} → b # α → Tmᴸ b α → Tmᴸ (sucᴮ b) (b ◅ α)
importTm b#α = renameTmᴸ (suc# b#α) (coerceᴺ (⊆-# b#α))

substTmᴸ : ∀ {α β b b₂} → b₂ # β → (Name α → Tmᴸ b₂ β) → Tmᴸ b α → Tmᴸ b₂ β
substTmᴸ b₂#β f = TraverseTmᴸ.trTmᴸ (substKitᴸ V importTm) (f , b₂#β)

substTmᴸ0 : ∀ {α b} → (Name α → Tmᴸ (0 ᴮ) ø) → Tmᴸ b α → Tmᴸ (0 ᴮ) ø
substTmᴸ0 = substTmᴸ (0 ᴮ #ø)

substTm◅… : ∀ {α b} n → (Name α → Tmᴸ (n ᴮ) (n ◅…′ ø)) → Tmᴸ b α → Tmᴸ (n ᴮ) (n ◅…′ ø)
substTm◅… n = substTmᴸ (#-◅…′ø n)

_[0≔_] : Tmᴸ (1 ᴮ) (0 ᴮ ◅ ø) → Tmᴸ (0 ᴮ) ø → Tmᴸ (0 ᴮ) ø
t [0≔ u ] = substTmᴸ0 (exportWith u V) t

β-redᴸ : ∀ {b α} → b # α → Tmᴸ b α → Tmᴸ b α
β-redᴸ b# (ƛ fct · arg) = substTmᴸ b# (exportWith arg V) fct
β-redᴸ _  t             = t

-- While we can use the general technique we used for the nominal style
-- we can also take advantage of this canonical representation.
-- We do so in the EqTmᴸ module below.
-- Note however that while EqTmᴸ is more efficient and simpler the functions
-- in this module CmpTmᴸ are more general (does not require the same starting binder).
module CmpTmᴸ where
  cmpTmᴸ : ∀ {α β b₁ b₂} → |Cmp| Name α β → Tmᴸ b₁ α → Tmᴸ b₂ β → Bool
  cmpTmᴸ Δ (V x₁)      (V x₂)      = Δ x₁ x₂
  cmpTmᴸ Δ (t₁ · u₁)   (t₂ · u₂)   = cmpTmᴸ Δ t₁ t₂ ∧ cmpTmᴸ Δ u₁ u₂
  cmpTmᴸ Δ (ƛ t₁)     (ƛ t₂)      = cmpTmᴸ (extendNameCmp Δ) t₁ t₂
  cmpTmᴸ Δ (Let t₁ u₁) (Let t₂ u₂) = cmpTmᴸ Δ t₁ t₂ ∧ cmpTmᴸ (extendNameCmp Δ) u₁ u₂
  cmpTmᴸ _ (` κ₁)      (` κ₂)      = κ₁ ==ᶜ κ₂
  cmpTmᴸ _ _ _ = false

  _==Tmᴸ_ : ∀ {α b₁ b₂} → Tmᴸ b₁ α → Tmᴸ b₂ α → Bool
  _==Tmᴸ_ = cmpTmᴸ _==ᴺ_

module EqTmᴸ where
  _==Tmᴸ_ : ∀ {α b} → Tmᴸ b α → Tmᴸ b α → Bool
  V x₁      ==Tmᴸ V x₂      = x₁ ==ᴺ x₂
  (t₁ · u₁) ==Tmᴸ (t₂ · u₂)  = t₁ ==Tmᴸ t₂ ∧ u₁ ==Tmᴸ u₂
  ƛ t₁      ==Tmᴸ ƛ t₂      = t₁ ==Tmᴸ t₂
  Let t₁ u₁ ==Tmᴸ Let t₂ u₂  = t₁ ==Tmᴸ t₂ ∧ u₁ ==Tmᴸ u₂
  ` κ₁      ==Tmᴸ ` κ₂      = κ₁ ==ᶜ κ₂
  _         ==Tmᴸ _         = false

import      Category.Monad.Partiality.NP as Pa
open        Pa using (_⊥; run_for_steps)
open import Coinduction
module KAM where -- Krivine Abstract Machine
  open Pa using (now; later; never)

  Tmø = Tmᴸ (0 ᴮ) ø

  Stack = List Tmø

  _★_ : Tmø → Stack → Tmø ⊥
  (t · u)   ★ π        = t ★ (u ∷ π)                -- push
  (ƛ t)     ★ (u ∷ π)  = later (♯ (t [0≔ u ] ★ π)) -- grab
  (Let t u) ★ π        = later (♯ (u [0≔ t ] ★ π))
  v         ★ []       = now v
  (V x)     ★ _        = Nameø-elim x
  (` _)     ★ _        = never

module Tmᴿ⇒Tmᴸ where
  record Env b α : Set where
    constructor _,_
    field
      trˢ : String → Name α
      b#  : b # α
  extEnv : ∀ {b α} → String → Env b α → Env (sucᴮ b) (b ◅ α)
  extEnv bˢ (trˢ , b#) = (trˢ′ , suc# b#)
    where trˢ′ = λ s → if bˢ == s then nameᴮ _ else coerceᴺ (⊆-# b#) (trˢ s)

  open Env
  ! : ∀ {α b} → Env b α → Tmᴿ → Tmᴸ b α
  ! Δ (V x) = V (trˢ Δ x)
  ! Δ (ƛ bˢ t) = ƛ (! (extEnv bˢ Δ) t)
  ! Δ (t · u) = ! Δ t · ! Δ u

  conv : ∀ {α} k → (String → Name α) → (k ᴮ) # α → Tmᴿ → Tmᴸ (k ᴮ) α
  conv _ trˢ k# = ! (trˢ , k#)

module Parser where
  open Tmᴿ⇒Tmᴸ using (conv)

  convø? : Tmᴿ →? Tmᴸ (0 ᴮ) ø
  convø? = map? (rerootTmᴸ 1 0) ∘ closeTmᴸ? ∘ conv 1 (const (0 ᴺ)) (suc# (0 ᴮ #ø))

  conv? : Tmᴿ →? (∀ {α} n → Tmᴸ (n ᴮ) α)
  conv? = map? (λ t {η} n → shiftøTmᴸ n (rerootTmᴸ 1 0 t)) ∘ closeTmᴸ? ∘ conv 1 (const (0 ᴺ)) (suc# (0 ᴮ #ø))

  open ParseConv convø? public renaming (parseConv? to parseTmᴸø?; parseConv to parseTmᴸø; TmOk to TmᴸøOk)
  open ParseConv conv? public renaming (parseConv? to parseTmᴸ?; parseConv to parseTmᴸ; TmOk to TmᴸOk)

open Parser public using (parseTmᴸ?; parseTmᴸ; TmᴸOk; parseTmᴸø?; parseTmᴸø; TmᴸøOk)

module Tmᴸ⇒Tmᴺ where
  ⟪_⟫ : ∀ {α b} → Tmᴸ b α → Tmᴺ α
  ⟪ V x     ⟫ = V x
  ⟪ ƛ t     ⟫ = ƛ _ ⟪ t ⟫
  ⟪ t · u   ⟫ = ⟪ t ⟫ · ⟪ u ⟫
  ⟪ Let t u ⟫ = ƛ _ ⟪ u ⟫ · ⟪ t ⟫
  ⟪ ` κ     ⟫ = `(Cst⇒ℕ κ)
open Tmᴸ⇒Tmᴺ public renaming (⟪_⟫ to convTmᴸ⇒Tmᴺ)

showTmᴸ : ∀ {b} → Tmᴸ b ø → String
showTmᴸ = showTmᴺ ∘ convTmᴸ⇒Tmᴺ

module TraverseTmᴺ⇒Tmᴸ {E}   (E-app : Applicative E)
                       {Env : World → Binder → World → Set}
                       (trName : ∀ {α b β} → Env α b β → Name α → E (Tmᴸ b β))
                       (extEnv : ∀ {α b₂ β} b₁ (Δ : Env α b₂ β) → Env (b₁ ◅ α) (sucᴮ b₂) (b₂ ◅ β)) where

  open Applicative E-app

  tr : ∀ {α b β} → Env α b β → Tmᴺ α → E (Tmᴸ b β)
  tr Δ (V x)       = trName Δ x
  tr Δ (t · u)     = pure _·_ ⊛ tr Δ t ⊛ tr Δ u
  tr Δ (ƛ b t)     = pure ƛ ⊛ tr (extEnv b Δ) t
  tr Δ (Let b t u) = pure Let ⊛ tr Δ t ⊛ tr (extEnv b Δ) u
  tr _ (` κ)       = pure $ `(ℕ⇒Cst κ)

module ConvTmᴺ⇒Tmᴸ where
  Env : World → Binder → World → Set
  Env α b β = RenameEnvᴸ b b α β
  open SubstEnvᴸ
  extEnv : ∀ {α b₂ β} b₁ (Δ : Env α b₂ β) → Env (b₁ ◅ α) (sucᴮ b₂) (b₂ ◅ β)
  extEnv _ = Renameᴸ.extEnvᴸ
  open TraverseTmᴺ⇒Tmᴸ id-app (λ Δ → V ∘′ trNameᴸ Δ) extEnv -- renaming (tr to convTmᴺ⇒Tmᴸ)

  convTmᴺ⇒Tmᴸ : ∀ {α b β} → b # β → (Name α → Name β) → Tmᴺ α → Tmᴸ b β
  convTmᴺ⇒Tmᴸ b#β trName = tr (trName , b#β)

  convTmᴺø⇒Tmᴸ : ∀ {b β} → b # β → Tmᴺ ø → Tmᴸ b β
  convTmᴺø⇒Tmᴸ b#β = convTmᴺ⇒Tmᴸ b#β Nameø-elim

  convTmᴺø⇒Tmᴸø : ∀ b → Tmᴺ ø → Tmᴸ b ø
  convTmᴺø⇒Tmᴸø b = convTmᴺø⇒Tmᴸ (b #ø)
