module Data.Fin.Abs (W : Set) (ø : W) (_^¹ _⁺¹ : W → W) {-; z≤n; s≤s)-} where

open import Data.Nat as Nat
  using (ℕ; zero; suc)
  renaming ( _+_ to _N+_; _∸_ to _N∸_
           ; _≤_ to _N≤_; _<_ to _N<_; _≤?_ to _N≤?_)

open import Function
open import Level hiding (lift)
open import Relation.Nullary.Decidable
open import Relation.Binary

------------------------------------------------------------------------
-- Types

-- Fin (ø ^ n) is a type with n elements.

data Fin : W → Set where
  zero : {n : W} → Fin (n ^¹)
  suc  : {n : W} (i : Fin n) → Fin (n ⁺¹)

-- A conversion: toℕ "n" = n.

toℕ : ∀ {n} → Fin n → ℕ
toℕ zero    = 0
toℕ (suc i) = suc (toℕ i)

toW : ∀ {n} → Fin n → W
toW zero    = ø
toW (suc i) = toW i ^¹

_^_ : W → ℕ → W
w ^ zero  = w
w ^ suc n = (w ^ n)^¹

_+w_ : W → ℕ → W
w +w zero  = w
w +w suc n = (w +w n)⁺¹

-- A Fin-indexed variant of Fin.

Fin′ : ∀ {n} → Fin n → Set
Fin′ i = Fin (toW i)

------------------------------------------------------------------------
-- Conversions

-- toℕ is defined above.

-- makes no sense in Abs
fromℕ : (n : ℕ) → Fin (ø ^ suc n)
fromℕ zero    = zero
fromℕ (suc n) = {!suc (fromℕ n)!}

-- fromℕ≤ {m} _ = "m".

fromℕ≤ : ∀ {m n} → m N< n → Fin (ø ^ n)
fromℕ≤ (Nat.s≤s Nat.z≤n)       = zero
fromℕ≤ (Nat.s≤s (Nat.s≤s m≤n)) = {!suc (fromℕ≤ (Nat.s≤s m≤n))!}

-- # m = "m".

#_ : ∀ m {n} {m<n : True (suc m N≤? n)} → Fin (ø ^ n)
#_ _ {m<n = m<n} = fromℕ≤ (toWitness m<n)

inject₁ : ∀ {m} → Fin m → Fin (m ^¹)
inject₁ zero    = zero
inject₁ (suc i) = {!suc (inject₁ i)!}

------------------------------------------------------------------------
-- Operations

-- Fold.

fold : ∀ (T : W → Set) {m} →
       (∀ {n} → T n → T (n ⁺¹)) →
       (∀ {n} → T (n ^¹)) →
       Fin m → T m
fold T f x zero    = x
fold T f x (suc i) = f (fold T f x i)

-- "m" + "n" = "m + n".

infixl 6 _+_

_+_ : ∀ {m n} (i : Fin m) (j : Fin n) → Fin (n +w toℕ i)
zero  + j = j
suc i + j = suc (i + j)
