{-# OPTIONS --universe-polymorphism #-}

open import Level as L using (_⊔_)
open import Function
open import Data.Product.Extras
open import Data.Sum
open import Data.Bool using (Bool)
open import Data.Unit
open import Data.Empty

module Data.Indexed {i} {Ix : Set i} where

|Set| : ∀ ℓ → Set _
|Set| ℓ = Ix → Set ℓ

|∀| : ∀ {f} (F : |Set| f) → Set _
|∀| F = ∀ {x} → F x

|∃| : ∀ {f} (F : |Set| f) → Set _
|∃| F = ∃[ x ] F x

|pure| : ∀ {a} (A : Set a) → |Set| _
|pure| A _ = A

-- an alias for |pure|
|K| : ∀ {a} (A : Set a) → |Set| _
|K| A _ = A

|Cmp| : (F : |Set| _) (i j : Ix) → Set
|Cmp| F i j = F i → F j → Bool

-- this version used to work (type-checking its uses) better than that
infixr 0 _|→|_
_|→|_ : ∀ {f g} (F : |Set| f) (G : |Set| g) → |Set| (f ⊔ g)
(F |→| G) i = F i → G i

infixr 0 _|$|_
_|$|_ : ∀ {f g} {F : |Set| f} {G : |Set| g} → |∀| ((F |→| G) |→| F |→| G)
_|$|_ x = x

infixr 2 _|×|_
_|×|_ : ∀ {f g} (F : |Set| f) (G : |Set| g) → |Set| _
(F |×| G) i = F i × G i

_|,|_ : ∀ {f g} {F : |Set| f} {G : |Set| g} → |∀| (F |→| G |→| F |×| G)
_|,|_ = _,_

pack : ∀ {f} {F : |Set| f} → (x : Ix) → F x → |∃| F
pack = _,_

infixr 1 _|⊎|_

_|⊎|_ : ∀ {f g} (F : |Set| f) (G : |Set| g) → |Set| _
(F |⊎| G) i = F i ⊎ G i

|inj₁| : ∀ {f g} {F : |Set| f} {G : |Set| g} → |∀|(F |→| F |⊎| G)
|inj₁| = inj₁

|inj₂| : ∀ {f g} {F : |Set| f} {G : |Set| g} → |∀|(G |→| F |⊎| G)
|inj₂| = inj₂

|⊤| : |Set| _
|⊤| _ = ⊤

|⊥| : |Set| _
|⊥| _ = ⊥

|¬| : ∀ {ℓ} → |Set| ℓ → |Set| _
-- Agda was accepting it previously
-- |¬| F = F |→| |⊥|
|¬| F = λ A → F A → ⊥

infixr 0 _|↦|_
_|↦|_ : ∀ {f g} (F : |Set| f) (G : |Set| g) → Set _
(F |↦| G) = |∀|(F |→| G)

|id| : ∀ {f} {F : |Set| f} → F |↦| F
|id| = id

_|∘|_ : ∀ {f g h} {F : |Set| f} {G : |Set| g} {H : |Set| h} → G |↦| H → F |↦| G → F |↦| H
f |∘| g = f ∘ g
