open import Data.Maybe.NP as Maybe
open import Data.Nat
open import Data.Unit
open import Data.Char renaming (_==_ to _==ᶜ_)
open import Data.String as String
open import Data.List as L using (List; []; _∷_; null; filter)
open import Data.Product.Extras hiding (map)
open import Data.Bool
open import Function
open import Coinduction
import Level as L
import Category.Monad as Cat
import Category.Monad.Partiality.NP as Pa
open   Pa using (_⊥; run_for_steps?; now; later; never; _∘⊥_; module M⊥)

module Text.Parser.Partial where

open M? L.zero using () renaming (_=<<_ to _=<<?_; _⊛_ to _⊛?_; join to join?)
open M⊥ L.zero using () renaming (_=<<_ to _=<<⊥_; _⊛_ to _⊛⊥_; _>>=_ to _>>=⊥_; _<$>_ to map⊥)

Parser⊥ : Set → Set
Parser⊥ A = List Char → (Maybe (A × List Char))⊥

pure : ∀ {A} → A → Parser⊥ A
pure a s = now (just (a , s))

infixr 3 _⟨|⟩_
_⟨|⟩_ : ∀ {A} → (f g : Parser⊥ A) → Parser⊥ A
_⟨|⟩_ f g s = maybe′ (now ∘ just) (g s) =<<⊥ f s

empty : ∀ {A} → Parser⊥ A
empty _ = now nothing

sat : (Char → Bool) → Parser⊥ Char
sat pred (x ∷ xs) = (if pred x then pure x else empty) xs
sat _    []       = now nothing

eof : Parser⊥ ⊤
eof s = if null s then pure _ [] else empty []

parser∞ : ∀ {A} → ∞(Parser⊥ A) → Parser⊥ A
parser∞ p s = later (♯ (♭ p) s)

parser⊥ : ∀ {A} → (Parser⊥ A)⊥ → Parser⊥ A
parser⊥ (now p)   s = p s
parser⊥ (later p) s = later (♯ parser⊥ (♭ p) s)

-- The remaning input is dropped, you may want to first
-- combine your parser `p' with `eof': `p <* eof'
runParser : ∀ {A} → Parser⊥ A → List Char → (Maybe A)⊥
runParser p s = (map⊥ ∘ map?) proj₁ (p s)

infixr 0 _>>=_
_>>=_ : ∀ {A B} → Parser⊥ A → (A → Parser⊥ B) → Parser⊥ B
_>>=_ pa f = maybe′ (uncurry f) (now nothing) ∘⊥ pa

infixl 0 _=<<_
_=<<_ : ∀ {A B} → (A → Parser⊥ B) → Parser⊥ A → Parser⊥ B
_=<<_ f pa = pa >>= f

join : ∀ {A} → Parser⊥ (Parser⊥ A) → Parser⊥ A
join = _=<<_ id

map : ∀ {A B} → (A → B) → Parser⊥ A → Parser⊥ B
map f pa = pa >>= pure ∘ f

infixl 4 _⊛_ _*>_ _<*_
_⊛_ : ∀ {A B} → Parser⊥ (A → B) → Parser⊥ A → Parser⊥ B
_⊛_ pf pa = pf >>= λ f → pa >>= λ a → pure (f a)

_*>_ : ∀ {A B} → Parser⊥ A → Parser⊥ B → Parser⊥ B
p₁ *> p₂ = p₁ >>= const p₂

_<*_ : ∀ {A B} → Parser⊥ A → Parser⊥ B → Parser⊥ A
p₁ <* p₂ = p₁ >>= λ x → p₂ *> pure x

choices : ∀ {A} → List (Parser⊥ A) → Parser⊥ A
choices = L.foldr _⟨|⟩_ empty

{-
mutual
  many : ∀ {A} → Parser⊥ A → Parser⊥ (List A)
  many p = some p ⟨|⟩ pure []

  some : ∀ {A} → Parser⊥ A → Parser⊥ (List A)
  some p = pure _∷_ ⊛ p ⊛ many p
-}

char : Char → Parser⊥ ⊤
char c = sat (_==ᶜ_ c) *> pure _

private
  notElem : Char → List Char → Bool
  notElem c = null ∘ filter (_==ᶜ_ c)

  syntax notElem c cs = c `notElem` cs

  elem : Char → List Char → Bool
  elem c = not ∘ notElem c

  syntax elem c cs = c `elem` cs

oneOf : List Char → Parser⊥ Char
oneOf cs = sat (λ c → c `elem` cs)

noneOf : List Char → Parser⊥ Char
noneOf cs = sat (λ c → c `notElem` cs)
